//
//  TextStepperField.swift
//  CustomControls
//
//  Created by Evan Owen on 5/23/16.
//  Copyright © 2016 Evan Owen. All rights reserved.
//

import Cocoa

class TextStepperField: NSTextField, NSTextFieldDelegate {

    var stepper: NSStepper!
    
    var stepperWidth: CGFloat {
        switch controlSize {
        case .RegularControlSize:
            return 16.0
        case .SmallControlSize:
            return 14.0
        case .MiniControlSize:
            return 12.0
        }
    }
    var smallIncrement = 1 {
        didSet {
            stepper.increment = Double(smallIncrement)
        }
    }
    var bigIncrement = 10
    
    override init(frame frameRect: NSRect) {
        super.init(frame: frameRect)
        initStepper()
    }
    
    required init?(coder: NSCoder) {
        super.init(coder: coder)
    }
    
    override func awakeFromNib() {
        initStepper()
    }
    
    func initStepper() {
        
        // TODO: right to left reading order
        if userInterfaceLayoutDirection == .LeftToRight {
            
        } else {
            
        }
        
        integerValue = 0
        delegate = self

        let stepperFrame = NSMakeRect(frame.width - stepperWidth, 0, stepperWidth, frame.height)
        
        stepper = NSStepper(frame: stepperFrame)
        stepper.integerValue = integerValue
        stepper.bind("hidden", toObject: self, withKeyPath: "hidden", options: nil)
        stepper.bind("controlSize", toObject: self, withKeyPath: "controlSize", options: nil)
        stepper.bind("enabled", toObject: self, withKeyPath: "enabled", options: nil)
        if let numberFormatter = self.formatter as? NSNumberFormatter, let max = numberFormatter.maximum, let min = numberFormatter.minimum {
            stepper.maxValue = Double(max)
            stepper.minValue = Double(min)
        }
        stepper.target = self
        stepper.action = #selector(TextStepperField.stepperAction(_:))
        stepper.valueWraps = false
        stepper.autorepeat = true
        stepper.increment = Double(smallIncrement)
        
        addSubview(stepper)
        if let bindingInfo = infoForBinding(NSValueBinding),
            let observedObject = bindingInfo[NSObservedObjectKey],
            let observedKeyPath = bindingInfo[NSObservedKeyPathKey] as? String {
            stepper.bind(NSValueBinding, toObject: observedObject, withKeyPath: observedKeyPath, options: nil)
//        } else {
        }
    }
    
    func stepperAction(sender: AnyObject) {
        window?.makeFirstResponder(self)
        setStepperValue(stepper.integerValue)
    }
    
    func setStepperValue(integerValue: Int) {
        
        // sync stepper and text field values
        let previousValue = self.integerValue
        stepper.objectValue = integerValue
        objectValue = integerValue
        
        // return if the value didn't change
        if self.integerValue == previousValue { return }
        
        // update binding, if valid
        if let bindingInfo = infoForBinding(NSValueBinding), let observedKeyPath = bindingInfo[NSObservedKeyPathKey] as? String {
            bindingInfo[NSObservedObjectKey]?.setValue(objectValue, forKeyPath: observedKeyPath)
        }
    }
    
    override var acceptsFirstResponder: Bool { return true }
    
    override func resetCursorRects() {
        let cursorRect = NSMakeRect(0, 0, frame.width - stepperWidth, frame.height)
        addCursorRect(cursorRect, cursor: NSCursor.IBeamCursor())
    }
    
    override func resizeSubviewsWithOldSize(oldSize: NSSize) {
        stepper.frame = NSMakeRect(frame.width - stepperWidth, 0, stepperWidth, frame.height)
    }
    
    // TODO: resize, menu
    
    func control(control: NSControl, textView: NSTextView, doCommandBySelector commandSelector: Selector) -> Bool {
        
        // TODO: get the stepper button to highlight
        
        switch commandSelector {
        case #selector(NSResponder.moveUp):
//            integerValue += smallIncrement
            setStepperValue(integerValue + smallIncrement)
//            stepper.intValue = intValue
            return true

        case #selector(NSResponder.moveDown):
//            integerValue -= smallIncrement
            setStepperValue(integerValue - smallIncrement)
//            stepper.intValue = intValue
            return true

        case #selector(NSResponder.moveToBeginningOfDocument(_:)):
//            integerValue += bigIncrement
            setStepperValue(integerValue + bigIncrement)
//            stepper.intValue = intValue
            return true

        case #selector(NSResponder.moveToEndOfDocument(_:)):
//            integerValue -= bigIncrement
            setStepperValue(integerValue - bigIncrement)
//            stepper.intValue = intValue
            return true

        default:
            return false
        }
    }
}
