//
//  Document.swift
//  FF6Tools
//
//  Created by Evan Owen on 8/18/16.
//  Copyright © 2016 Evan Owen. All rights reserved.
//

import Cocoa

class Document: NSDocument {

    var rom: FF6ROM!
    var romWindowController: ROMWindowController?

    override init() {
        super.init()
        // Add your subclass-specific initialization here.
    }

    override func makeWindowControllers() {
        romWindowController = FF6WindowController(windowNibName: NSNib.Name(rawValue: "Document"))
        romWindowController!.rom = rom
        addWindowController(romWindowController!)
    }
    
    override class var autosavesInPlace: Bool {
        return false
    }

    override func writeSafely(to url: URL, ofType typeName: String, for saveOperation: NSDocument.SaveOperationType) throws {
        
        // assemble and write the ROM to file
        romWindowController?.stopObservingAllObjects()
        try rom.write(to: url)
        romWindowController?.startObservingAllObjects()
        
        // create the ROM definition and write it to file
        let definition = rom.createDefinition()
        var definitionURL = url.deletingPathExtension()
        definitionURL = definitionURL.appendingPathExtension("plist")
        if definition.write(to: definitionURL, atomically: true) == false {
            throw NSError(domain: "Unable to save ROM definition.", code: 0, userInfo: nil)
        }
    }
    
    override func read(from url: URL, ofType typeName: String) throws {
        rom = FF6ROM(contentsOfURL: url)
        if rom == nil {
            throw NSError(domain: "Invalid ROM", code: 0, userInfo: nil)
        }
    }
    
    override func revert(toContentsOf url: URL, ofType typeName: String) throws {

        try read(from: url, ofType: typeName)

        romWindowController?.deselectAllObjects()
        romWindowController?.rom = self.rom
        romWindowController?.navigatorController?.navigatorView.reloadData()
        romWindowController?.navigatorController?.navigatorView.collapseItem(nil, collapseChildren: true)
        romWindowController?.listView?.tabViewItem(at: 0).view = NSView()
        romWindowController?.listView?.selectTabViewItem(at: 0)
        romWindowController?.editorView?.tabViewItem(at: 0).view = NSView()
        romWindowController?.editorView?.selectTabViewItem(at: 0)
        romWindowController?.toolboxView?.tabViewItem(at: 0).view = NSView()
        romWindowController?.toolboxView?.selectTabViewItem(at: 0)
        undoManager?.removeAllActions()
        updateChangeCount(NSDocument.ChangeType.changeCleared)
    }

}

