//
//  FF6BattleEditor.swift
//  FF6Tools
//
//  Created by Evan Owen on 2/6/16.
//  Copyright © 2016 Evan Owen. All rights reserved.
//

import Cocoa

class FF6BattleEditor: FF6BattleView, NSMenuDelegate {
    
    @IBOutlet weak var vramView: FF6BattleVRAMView?
    
    override var acceptsFirstResponder: Bool { return true }

    @IBAction func changeBattleType(sender: AnyObject) {
        guard let menuItem = sender as? NSMenuItem else { return }
        guard let type = BattleType(rawValue: menuItem.tag) else { return }
        
        let disableNormal = auxBattleProperties.boolValue(forKey: "disableNormal")
        let disableBack = auxBattleProperties.boolValue(forKey: "disableBack")
        let disablePincer = auxBattleProperties.boolValue(forKey: "disablePincer")
        let disableSide = auxBattleProperties.boolValue(forKey: "disableSide")
        
        if type == .normal && disableNormal {
            return
        } else if type == .back && disableBack {
            return
        } else if type == .pincer && disablePincer {
            return
        } else if type == .side && disableSide {
            return
        }

        self.battleType = type
        needsDisplay = true
    }
    
    @IBAction func changeBackground(sender: AnyObject) {
        guard let menuItem = sender as? NSMenuItem else { return }
        
        background = menuItem.tag
    }
    
    override func mouseDown(with theEvent: NSEvent) {
        var event = theEvent
        
        // get the point that was clicked
        let clickedPoint = convert(event.locationInWindow, from: nil)
        if !bounds.contains(clickedPoint) { return }
        
        if let clickedMonster = monsterAtPoint(clickedPoint) {
            
            if theEvent.clickCount == 2 && clickedMonster === selectedObject {
                // double clicked a monster - select monster properties
                windowController?.deselectAllObjects()
                let monsterProperties = rom.value(at: clickedMonster.index, inPropertyWithKey: "monsterProperties") as? ROMAssembly
                windowController.selectObject(monsterProperties)
                return
            } else if theEvent.clickCount == 3 && clickedMonster === selectedObject {
                // triple clicked a monster - select monster script
                let scriptPointer = rom.value(at: clickedMonster.index, inPropertyWithKey: "monsterScriptPointers") as? ROMData
                let reference = scriptPointer?.value(forKey: "scriptPointer") as? ROMReference
                let referencedCommand = reference?.referencedAssembly as? ROMCommand
                windowController?.deselectAllObjects()
                windowController?.selectObject(referencedCommand?.script, identifierOverride: nil)
                windowController?.scriptController?.script = referencedCommand?.script
                windowController?.scriptController?.select(command: referencedCommand)
                return
            }
            
            // selected a monster
            selectedObject = clickedMonster
            vramView?.selectedSlot = clickedMonster.slot
            windowController?.deselectAllObjects()
            windowController?.selectObject(clickedMonster, identifierOverride: nil)
            
            // save the monster's original position
            let oldX = clickedMonster.x
            let oldY = clickedMonster.y
            
            var monsterHasNotMoved = true
            while true {
                
                switch event.type {
                case .leftMouseUp:
                    return
                    
                case .leftMouseDragged, .leftMouseDown:
                    
                    // get the dragged/clicked point
                    let draggedPoint = convert(event.locationInWindow, from: nil)
                    
                    // get the delta from the original clicked point
                    let deltaX = (draggedPoint.x - clickedPoint.x) / zoom / 8.0
                    let deltaY = (draggedPoint.y - clickedPoint.y) / zoom / 8.0
                    
                    // limit the position to valid values
                    let newX = min(15, max(oldX + Int(deltaX), 0))
                    let newY = min(15, max(oldY - Int(deltaY), 0))
                    
                    // break if the monster didn't move
                    if ((newX == clickedMonster.x) && (newY == clickedMonster.y)) { break }
                    
                    // set the new position
                    setMonsterPosition(clickedMonster.battleProperties, slot: clickedMonster.slot, x: newX, y: newY, canUndo: monsterHasNotMoved)
                    monsterHasNotMoved = false
                    
                default:
                    break
                }
                event = window!.nextEvent(matching: [.leftMouseUp, .leftMouseDragged])!
            }
            
        } else if let clickedCharacter = characterAtPoint(clickedPoint), clickedCharacter.hasAI {
            // selected a character
            selectedObject = clickedCharacter
            vramView?.selectedSlot = clickedCharacter.slot + 6
            windowController?.deselectAllObjects()
            windowController?.selectObject(clickedCharacter, identifierOverride: nil)
            
            // save the character's original position
            let oldX = clickedCharacter.x
            let oldY = clickedCharacter.y
            
            var characterHasNotMoved = true
            while true {
                
                switch event.type {
                case .leftMouseUp:
                    return
                    
                case .leftMouseDragged, .leftMouseDown:
                    
                    // get the dragged/clicked point
                    let draggedPoint = convert(event.locationInWindow, from: nil)
                    
                    // get the delta from the original clicked point
                    let deltaX = (draggedPoint.x - clickedPoint.x) / zoom / 2.0
                    let deltaY = (draggedPoint.y - clickedPoint.y) / zoom / 2.0
                    
                    // limit the position to valid values
                    let newX = min(255, max(oldX + Int(deltaX), 0))
                    let newY = min(255, max(oldY - Int(deltaY), 0))
                    
                    // break if the monster didn't move
                    if ((newX == clickedCharacter.x) && (newY == clickedCharacter.y)) { break }
                    
                    // set the new position
                    setCharacterPosition(clickedCharacter.auxBattleProperties, slot: clickedCharacter.slot, x: newX, y: newY, canUndo: characterHasNotMoved)
                    characterHasNotMoved = false
                    
                default:
                    break
                }
                event = window!.nextEvent(matching: [.leftMouseUp, .leftMouseDragged])!
            }
            
        } else if !windowController.objectIsSelected(battleProperties) {
            // selected nothing
            selectedObject = nil
            vramView?.selectedSlot = nil
            windowController.deselectAllObjects()
            windowController.selectObject(battleProperties)
//            windowController?.selectObject(atIndex: battle, inArray: rom.value(forKey: "battleProperties") as? ROMArray)
        }
    }
    
//    override func rightMouseDown(with theEvent: NSEvent) {
//        
//        if menu != nil {
//            menu!.update()
//            NSMenu.popUpContextMenu(menu!, with: theEvent, for: self)
//        }
//        
//        // right click cycles through the battle types (for now)
////        switch battleType {
////        case .normal:   battleType = .back
////        case .back:     battleType = .pincer
////        case .pincer:   battleType = .side
////        case .side:     battleType = .normal
////        }
////        needsDisplay = true
//    }
    
//    override func keyDown(with theEvent: NSEvent) {
//        // arrow keys cycle through battle backgrounds
//        let s   =   theEvent.charactersIgnoringModifiers!
//        let s1  =   s.unicodeScalars
//        let s2  =   s1[s1.startIndex].value
//        let s3  =   Int(s2)
//        switch s3 {
//        case NSUpArrowFunctionKey:    background += 1 // next background
//        case NSDownArrowFunctionKey:  background -= 1 // previous background
//        default: return
//        }
//    }
    
    func deselectAll() {
        selectedObject = nil
        vramView?.selectedSlot = nil
    }

    func selectMonsterSlot(_ slot: Int) {
        if let monster = monsterInSlot(slot) {
            selectedObject = monster
            vramView?.selectedSlot = slot
        }
    }

    @objc func setMonsterPosition(_ battleProperties: ROMData, slot: Int, x: Int, y: Int, canUndo: Bool) {
        
        let monster = FF6BattleMonster(rom: ff6ROM, battleProperties: battleProperties, slot: slot)
        
        // save undo data
        if canUndo {
            if let undo = windowController?.undoManager {
                let target = undo.prepare(withInvocationTarget: self) as AnyObject
                target.setMonsterPosition(battleProperties, slot: slot, x: monster.x, y: monster.y, canUndo: true)
                if (!undo.isUndoing) {
                    undo.setActionName("Move Monster")
                }
            }
        }
        
        // set the monster's position
        windowController?.stopObservingAllObjects()
        monster.x = x
        monster.y = y
        windowController?.startObservingAllObjects()
    }

    func selectCharacterSlot(_ slot: Int) {
        // select a monster slot
        for character in characters {
            if character.slot == slot {
                selectedObject = character
                vramView?.selectedSlot = slot + 6
            }
        }
    }

    @objc func setCharacterPosition(_ auxBattleProperties: ROMData, slot: Int, x: Int, y: Int, canUndo: Bool) {
        // create a dummy character
        let character = FF6BattleCharacter(rom: ff6ROM, auxBattleProperties: auxBattleProperties, slot: slot)
        
        // save undo data
        if canUndo {
            if let undo = windowController?.undoManager {
                let target = undo.prepare(withInvocationTarget: self) as AnyObject
                target.setCharacterPosition(auxBattleProperties, slot: slot, x: character.x, y: character.y, canUndo: true)
                if (!undo.isUndoing) {
                    undo.setActionName("Move Character")
                }
            }
        }
        
        // set the character's position
        windowController?.stopObservingAllObjects()
        character.x = x
        character.y = y
        windowController?.startObservingAllObjects()
    }

    func menuNeedsUpdate(_ menu: NSMenu) {
        if menu.title == "Battle Type" {
            
            let disabledTypes: [Bool] = [
                auxBattleProperties.boolValue(forKey: "disableNormal"),
                auxBattleProperties.boolValue(forKey: "disableBack"),
                auxBattleProperties.boolValue(forKey: "disablePincer"),
                auxBattleProperties.boolValue(forKey: "disableSide")
            ]
            
            for index in 0..<4 {
                menu.item(at: index)?.state = NSControl.StateValue(rawValue: index == battleType.rawValue ? 1 : 0)
                menu.item(at: index)?.action = disabledTypes[index] ? nil : #selector(FF6BattleEditor.changeBattleType(sender:))
            }
        } else if menu.title == "Background" {
            guard let battleBackgroundStringTable = rom.stringTable(forKey: "battleBackgrounds") else { return }
            
            var menuItems = [NSMenuItem]()
            for (key, object) in battleBackgroundStringTable.stringDictionary {
                guard let index = key as? Int, let backgroundString = object as? ROMString else { continue }
                if index >= 63 { continue }
                
                let menuItem = NSMenuItem(title: backgroundString.numberedString, action: #selector(FF6BattleEditor.changeBackground(sender:)), keyEquivalent: "")
                menuItem.tag = index
                menuItem.target = self
                if index == background { menuItem.state = NSControl.StateValue(rawValue: 1) }
                menuItems.append(menuItem)
            }
            
            menu.removeAllItems()
            menuItems.sort{ $0.tag < $1.tag }
            for menuItem in menuItems {
                menu.addItem(menuItem)
            }
        }
    }
}

class FF6MonsterView: FF6View {

    override func draw(_ dirtyRect: NSRect) {

        super.draw(dirtyRect)
        
        // fill the rect with orange
        NSColor.controlColor.set()
        NSBezierPath.fill(dirtyRect)

        guard let iref = ff6ROM.monsterImage(windowController.selectedIndex) else { return }

        let monsterSize = ff6ROM.size(ofMonster: windowController.selectedIndex)
        let monsterWidth = CGFloat(monsterSize.width * 8)
        let monsterHeight = CGFloat(monsterSize.height * 8)
        let zoom: CGFloat = 1.0

        // draw the image
        let imageHeight = CGFloat(iref.height) * zoom
        let imageWidth = CGFloat(iref.width) * zoom
        let image = NSImage(cgImage: iref, size: NSMakeSize(imageWidth, imageHeight)) // this is an NSImage scaled to the zoom factor
        let imageRect = NSMakeRect(0, imageHeight - monsterHeight * zoom, monsterWidth * zoom, monsterHeight * zoom)

        // center the image in the view
        let x = (bounds.width - monsterWidth * zoom) / 2
        let y = (bounds.height - monsterHeight * zoom) / 2
        let imagePoint = NSMakePoint(x, y)
        
        NSGraphicsContext.current?.imageInterpolation = imageInterpolation
        image.draw(at: imagePoint, from: imageRect, operation: NSCompositingOperation.sourceOver, fraction: 1.0)
    }
}
