//
//  FF6BattleVRAMView.swift
//  FF6Tools
//
//  Created by Evan Owen on 2/13/16.
//  Copyright © 2016 Evan Owen. All rights reserved.
//

import Cocoa

class FF6BattleVRAMView: FF6View, NSMenuDelegate {

//    @IBOutlet weak var windowController: ROMWindowController!
//    var rom: FF6ROM { return windowController.rom as! FF6ROM }
    
//    @IBOutlet weak var parentView: NSView?
    @IBOutlet weak var battleView: FF6BattleEditor?
//    var ff6ROM: FF6ROM { return rom as! FF6ROM }

    var battle = 0 {
        didSet {
            loadBattle()
            if battle != oldValue { selectedSlot = nil }
        }
    }
    fileprivate var battleProperties: ROMData!
    fileprivate var auxBattleProperties: ROMData!
    fileprivate var vramMap: ROMArray!
    
    var zoom: CGFloat { return max((bounds.width) / 128, (bounds.height) / 128) }
    var selectedSlot: Int? { didSet { needsDisplay = true } }
    
    deinit {
        // remove observer from the monster VRAM map index
        if battleProperties == nil { return }
        stopObserving(object: battleProperties)
//        NSUserDefaultsController.shared().removeObserver(self, forKeyPath: "values.FF6BackgroundColor")
//        NSUserDefaultsController.shared().removeObserver(self, forKeyPath: "values.FF6InterpolationMode")
    }
    
//    override func awakeFromNib() {
//        NSUserDefaultsController.shared().addObserver(self, forKeyPath: "values.FF6BackgroundColor", options: NSKeyValueObservingOptions.new, context: &FF6BattleVRAMView.context)
//        NSUserDefaultsController.shared().addObserver(self, forKeyPath: "values.FF6InterpolationMode", options: NSKeyValueObservingOptions.new, context: &FF6BattleVRAMView.context)
//    }
    
    func slotAtPoint(_ point: NSPoint) -> Int? {
        for slot in 1...10 {
            if rectForSlot(slot).contains(point) {
                return slot
            }
        }
        return nil
    }
    
    func rectForSlot(_ slot: Int) -> NSRect {
//        if slot > 6 {
//            // character slot
//            return NSMakeRect(128 * zoom, CGFloat(3 - (slot - 7)) * 32 * zoom, 16 * zoom, 32 * zoom)
//        }
        
        // monster slot, get vram map data
        if slot > vramMap.array.count { return NSZeroRect }
        guard let vramMapData = vramMap.array[slot - 1] as? ROMData else { return NSZeroRect }
        let vramAddress = vramMapData.intValue(forKey: "vramAddress")
        let slotWidth = vramMapData.intValue(forKey: "width")
        let slotHeight = vramMapData.intValue(forKey: "height")
        let x = (vramAddress & 0x01FF) >> 5
        let y = (vramAddress & 0xFE00) >> 9
        return NSMakeRect(CGFloat(x) * 8 * zoom, CGFloat(16 - y - slotHeight) * 8 * zoom, CGFloat(slotWidth * 8) * zoom, CGFloat(slotHeight * 8) * zoom)
    }
    
    override func mouseDown(with theEvent: NSEvent) {
        // get the point that was clicked
        let clickedPoint = convert(theEvent.locationInWindow, from: nil)
        if !bounds.contains(clickedPoint) { return }
        
        if let clickedSlot = slotAtPoint(clickedPoint) , clickedSlot != selectedSlot {
            selectedSlot = clickedSlot
            if clickedSlot <= 6 {
                windowController.deselectAllObjects()
                windowController.selectObject(battleView?.monsterInSlot(clickedSlot))
            } else {
                windowController.deselectAllObjects()
                windowController.selectObject(battleView?.characterInSlot(clickedSlot - 6))
            }
        }
    }
    
    override func rightMouseDown(with theEvent: NSEvent) {
        mouseDown(with: theEvent)
        super.rightMouseDown(with: theEvent)
    }
    
    func loadBattle() {
        if battleProperties != nil { stopObserving(object: battleProperties) }
        battleProperties = rom.mutableArrayValue(forKey: "battleProperties").object(at: battle) as? ROMData
        auxBattleProperties = rom.mutableArrayValue(forKey: "auxBattleProperties").object(at: battle) as? ROMData
        let vramMapIndex = battleProperties.intValue(forKey: "vramMap")
        vramMap = rom.mutableArrayValue(forKey: "monsterVRAMMaps").object(at: vramMapIndex) as? ROMArray
        startObserving(object: battleProperties)
        needsDisplay = true
    }
    
    override func draw(_ dirtyRect: NSRect) {
        super.draw(dirtyRect)

        // fill the view with the background color
        
//        var backgroundColor = NSColor.orange
//        if let userColorData = UserDefaults.standard.data(forKey: "FF6BackgroundColor"),
//            let userColor = NSUnarchiver.unarchiveObject(with: userColorData) as? NSColor {
//            backgroundColor = userColor
//        }
        backgroundColor.set()
        NSBezierPath.fill(dirtyRect)
        
        // draw each monster slot
        for slot in 1...6 {
            
            let slotRect = rectForSlot(slot)
            
            // skip slots that are not in the VRAM map
            if slotRect.isEmpty { continue }
            
            // draw a box around the slot
            NSColor.gridColor.set()
            NSBezierPath.stroke(slotRect)
            
            // draw the monster
            let monster = FF6BattleMonster(rom: ff6ROM, battleProperties: battleProperties, slot: slot)
            drawMonster(monster)

            // draw the slot number
            let style = NSMutableParagraphStyle()
            style.alignment = NSTextAlignment.center
            let font = NSFont.boldSystemFont(ofSize: 24.0 * zoom)
            let fontHeight = NSLayoutManager().defaultLineHeight(for: font)
            let textColor = monster.slot == selectedSlot ? NSColor.keyboardFocusIndicatorColor : NSColor.black
            let attributes: [NSAttributedStringKey : Any] =
                [NSAttributedStringKey.paragraphStyle : style,
                 NSAttributedStringKey.foregroundColor : textColor.withAlphaComponent(0.5),
                 NSAttributedStringKey.strokeColor : NSColor.white.withAlphaComponent(1.0),
                 NSAttributedStringKey.strokeWidth : -3.0,
                 NSAttributedStringKey.font : font]
            let textRect = slotRect.insetBy(dx: 0.0, dy: (slotRect.height - fontHeight) / 2)
            NSString(format: "%d", slot, monster.name).draw(in: textRect, withAttributes: attributes)
        }
        
        // draw each character slot
//        for slot in 1...4 {
//         
//            let slotRect = rectForSlot(slot + 6)
//            if slotRect.isEmpty { continue }
//
//            // draw a box around the slot
//            NSColor.gridColor().set()
//            NSBezierPath.strokeRect(slotRect)
//
//            // draw the character
//            let character = FF6BattleCharacter(rom: rom, auxBattleProperties: auxBattleProperties, slot: slot)
//            drawCharacter(character, inRect: slotRect)
//        }
    }
    
    func drawMonster(_ monster: FF6BattleMonster) {
        
        if monster.isNull { return }
        
        // create a monster image
        guard let iref = ff6ROM.monsterImage(monster.index) else {
            rom.log("Unable to draw monster")
            return
        }
        
        let monsterRect = rectForSlot(monster.slot)
        
        // draw the image
        let imageHeight = CGFloat(iref.height)
        let imageWidth = CGFloat(iref.width)
        let image = NSImage(cgImage: iref, size: NSMakeSize(imageWidth, imageHeight)) // this is an NSImage scaled to the zoom factor
        let imageRect = NSMakeRect(0, imageHeight - monsterRect.height / zoom, monsterRect.width / zoom, monsterRect.height / zoom)
        
        // tint the monster red if it doesn't fit in it's VRAM slot
        let monsterSize = monster.size
        if (CGFloat(monsterSize.width * 8) * zoom > monsterRect.width) || (CGFloat(monsterSize.height * 8) * zoom > monsterRect.height) {
            image.tintWithColor(NSColor.red)
        }

        // tint the monster if it is selected
        if monster.slot == selectedSlot {
            image.tintWithColor(NSColor.keyboardFocusIndicatorColor)
        }
        
        NSGraphicsContext.current?.imageInterpolation = imageInterpolation
        image.draw(in: monsterRect.insetBy(dx: 1.0, dy: 1.0), from: imageRect, operation: NSCompositingOperation.sourceOver, fraction: 1.0)
    }
    
//    func drawCharacter(_ character: FF6BattleCharacter, inRect rect: NSRect) {
//        
//        var graphics = character.slot - 1
//        if character.hasAI { graphics = character.graphics }
//        if graphics == 0xFF { graphics = character.character }
//        
//        // create a character image
//        guard let iref = rom.battleCharacterImage(graphics, frame: 0x11) else {
//            rom.log("Unable to create character image")
//            return
//        }
//        
//        // draw the image
//        let imageHeight = CGFloat(iref.height) * zoom
//        let imageWidth = CGFloat(iref.width) * zoom
//        let image = NSImage(cgImage: iref, size: NSMakeSize(imageWidth, imageHeight)) // this is an NSImage scaled to the zoom factor
//        let imageRect = NSMakeRect(0, imageHeight - rect.height, rect.width, rect.height)
//        
//        // tint the character if it is selected
//        if character.slot + 6 == selectedSlot {
//            image.tintWithColor(NSColor.keyboardFocusIndicatorColor)
//        }
//        
//        NSGraphicsContext.current()?.imageInterpolation = NSImageInterpolation.none
//        image.draw(in: rect, from: imageRect, operation: NSCompositingOperation.sourceOver, fraction: character.hidden ? 0.5 : 1.0)
//    }
    
    // MARK: observing methods
    
    func startObserving(object: AnyObject?) {
        
        guard let selected = object as? ROMSelectable else { return }
        
        // properties for ROMSelectable objects
        if let properties = selected.properties {
            for (key, _) in properties {
                guard let propertyIdentifier = key as? String else { continue }
                object?.addObserver(self, forKeyPath: propertyIdentifier, options: NSKeyValueObservingOptions.old, context: &FF6BattleVRAMView.context)
            }
        }
    }
    
    func stopObserving(object: AnyObject?) {
        
        guard let selected = object as? ROMSelectable else { return }
        
        // properties for ROMSelectable objects
        if let properties = selected.properties {
            for (key, _) in properties {
                guard let propertyIdentifier = key as? String else { continue }
                object?.removeObserver(self, forKeyPath: propertyIdentifier, context: &FF6BattleVRAMView.context)
            }
        }
    }

    override func observeValue(forKeyPath keyPath: String?, of object: Any?, change: [NSKeyValueChangeKey : Any]?, context: UnsafeMutableRawPointer?) {
        if context != &FF6BattleVRAMView.context {
            super.observeValue(forKeyPath: keyPath, of: object, change: change, context: context)
            return
        }

        // reload the battle data any time an observed object changes
        loadBattle()
    }
    
//    static var context = 0

    // MARK: NSMenuDelegate members
    
    func menuNeedsUpdate(_ menu: NSMenu) {
        
        if selectedSlot == nil {
            menu.item(at: 0)!.isEnabled = false
            menu.item(at: 1)!.isEnabled = true
            return
        }
        
        let monster = FF6BattleMonster(rom: ff6ROM, battleProperties: battleProperties, slot: selectedSlot!)
        if monster.isNull {
            menu.item(withTitle: "Insert Monster")?.isEnabled = true
            menu.item(withTitle: "Delete Monster")?.isEnabled = false
        } else {
            menu.item(withTitle: "Insert Monster")?.isEnabled = false
            menu.item(withTitle: "Delete Monster")?.isEnabled = true
        }
    }
    
//    @IBAction func insertMonster(sender: AnyObject) {
//        let monster = BattleMonster()
//        monster.rom = rom
//        monster.battleProperties = battleProperties
//        monster.slot = selectedSlot!
//        monster.index = 0
//        monster.present = true
//        monster.bg1 = false
//        monster.x = 8
//        monster.y = 8
//        needsDisplay = true
//        delegate?.monsterVRAMView(self, editedMonsterInSlot: selectedSlot!)
//    }
//    
//    @IBAction func deleteMonster(sender: AnyObject) {
//        let monster = BattleMonster()
//        monster.rom = rom
//        monster.battleProperties = battleProperties
//        monster.slot = selectedSlot!
//        monster.index = 0x01FF
//        monster.present = false
//        monster.bg1 = false
//        monster.x = 0
//        monster.y = 0
//        needsDisplay = true
//        delegate?.monsterVRAMView(self, editedMonsterInSlot: selectedSlot!)
//    }
}
