//
//  FF6MapTrigger.swift
//  FF6Tools
//
//  Created by Evan Owen on 7/26/16.
//
//

import Cocoa

class FF6MapTrigger: NSObject, ROMSelectable {
    
    // MARK: ROMSelectable properties
    
    var identifier: String? { return "triggerProperties" }
    var properties: NSDictionary? { return nil }
    var nib: String? { return nil }
    
    // this is a wrapper class for the trigger objects on the map
    
    let triggerArray: ROMArray
    let triggerAssembly: ROMData
    @objc var index: Int { return triggerArray.array.index(of: triggerAssembly) }
    var type: TriggerType
    
    @objc var isItem: Bool    { get { return triggerAssembly.intValue(forKey: "treasureType") == 4 } set { } }
    @objc var isGP: Bool      { get { return triggerAssembly.intValue(forKey: "treasureType") == 8 } set { } }
    @objc var isMonster: Bool { get { return triggerAssembly.intValue(forKey: "treasureType") == 2 } set { } }
    var x: Int          { return triggerAssembly.intValue(forKey: "x") }
    var y: Int          { return triggerAssembly.intValue(forKey: "y") }
    @objc var isLong: Bool    { get { return triggerAssembly.value(forKeyPath: "length") != nil } set { } }
    
    init(triggerArray: ROMArray, triggerAssembly: ROMData) {
        
        if triggerArray.array.index(of: triggerAssembly) == NSNotFound {
            Swift.print("Invalid Trigger")
        }
        
        self.triggerArray = triggerArray
        self.triggerAssembly = triggerAssembly
        self.type = TriggerType(triggerAssembly: triggerAssembly)
        super.init()
        
        triggerAssembly.addObserver(self, forKeyPath: "treasureType", options: NSKeyValueObservingOptions.prior, context: &FF6MapTrigger.context)
    }
    
    deinit {
        triggerAssembly.removeObserver(self, forKeyPath: "treasureType", context: &FF6MapTrigger.context)
    }
    
    static var context = 0
    
    override func observeValue(forKeyPath keyPath: String?, of object: Any?, change: [NSKeyValueChangeKey : Any]?, context: UnsafeMutableRawPointer?) {
        if context != &FF6MapTrigger.context {
            super.observeValue(forKeyPath: keyPath, of: object, change: change, context: context)
            return
        }
        
        if keyPath != "treasureType" { return }
        if change?[NSKeyValueChangeKey.notificationIsPriorKey] != nil {
            willChangeValue(forKey: "isItem")
            willChangeValue(forKey: "isGP")
            willChangeValue(forKey: "isMonster")
        } else {
            didChangeValue(forKey: "isItem")
            didChangeValue(forKey: "isGP")
            didChangeValue(forKey: "isMonster")
        }
    }
    
}

enum TriggerType: Int {
    case none = -1, npc = 0, longEntrance = 1, shortEntrance = 2, event = 3, treasure = 4, specialNPC = 5, worldEvent = 6
    
    init(triggerAssembly: ROMAssembly?) {
        
        if let vehicle = triggerAssembly?.value(forKey: "vehicle") as? Int, let specialGraphics = triggerAssembly?.value(forKey: "specialGraphics") as? Int {
            
            if vehicle == 0 && specialGraphics == 1 {
                // special NPC graphics
                self = .specialNPC
            } else {
                // normal NPC
                self = .npc
            }
        } else if triggerAssembly?.value(forKey: "length") != nil {
            self = .longEntrance
        } else if triggerAssembly?.value(forKey: "setParent") != nil {
            self = .shortEntrance
        } else if triggerAssembly?.value(forKey: "eventPointer") != nil {
            self = .event
        } else if triggerAssembly?.value(forKey: "treasureType") != nil {
            self = .treasure
        } else {
            self = .none
        }
    }
    
    var color: NSColor {
        switch self {
        case .none:          return NSColor.black
        case .npc:           return NSColor.white
        case .longEntrance:  return NSColor.green
        case .shortEntrance: return NSColor.red
        case .event:         return NSColor.blue
        case .worldEvent:    return NSColor.blue
        case .treasure:      return NSColor.yellow
        case .specialNPC:    return NSColor.cyan
        }
    }
    
    var arrayName: String? {
        switch self {
        case .none:          return nil
        case .npc:           return "npcProperties"
        case .longEntrance:  return "longEntranceTriggers"
        case .shortEntrance: return "shortEntranceTriggers"
        case .event:         return "eventTriggers"
        case .worldEvent:    return "eventTriggers"
        case .treasure:      return "treasureProperties"
        case .specialNPC:    return "npcProperties"
        }
    }
    
    var identifier: String {
        switch self {
        case .none:          return "noSelection"
        case .npc:           return "npc"
        case .longEntrance:  return "entrance"
        case .shortEntrance: return "entrance"
        case .event:         return "event"
        case .worldEvent:    return "worldEvent"
        case .treasure:      return "treasure"
        case .specialNPC:    return "specialNPC"
        }
    }
}
