//
//  ROMObject.swift
//  ROMTools
//
//  Created by Evan Owen on 8/18/16.
//  Copyright © 2016 Evan Owen. All rights reserved.
//

import Cocoa

protocol ROMSelectable: class {
    var identifier: String? { get }
    var properties: NSDictionary? { get }
    var nib: String? { get }
}

class ROMObject: NSObject, NSCopying, ROMSelectable {
    
    weak var rom: ROM!
    var identifier: String?
    @objc var properties: NSDictionary?
    var nib: String?
    var name: String?

    override init() {
        super.init()
    }
    
    init(_ rom: ROM, definition: NSDictionary? = nil) {
        self.rom = rom
        self.identifier = definition?["identifier"] as? String
        self.properties = definition?["properties"] as? NSDictionary
        self.nib = definition?["nib"] as? String
        self.name = definition?["name"] as? String
        super.init()
    }
    
    func createDefinition() -> NSMutableDictionary {
        let definition = NSMutableDictionary()
        definition["type"] = "ROMObject"
        definition["identifier"] = identifier
        definition["properties"] = properties
        definition["nib"] = nib
        definition["name"] = name
        return definition
    }
    
    func copy(with zone: NSZone?) -> Any {
        return ROMObject.createObject(rom, definition: createDefinition())
    }
    
    class func createObject(_ rom: ROM, definition: NSDictionary) -> ROMObject {
        guard let type = definition["type"] as? String else {
            // no assembly type found
            rom.log("Assembly type not found")
            return ROMObject(rom, definition: definition)
        }
        
        switch type {
        case "ROMAssembly":         return ROMAssembly(rom, definition: definition)
        case "ROMArray":            return ROMArray(rom, definition: definition)
        case "ROMData":             return ROMData(rom, definition: definition)
        case "ROMGraphics":         return ROMGraphics(rom, definition: definition)
        case "ROMPointerTable":     return ROMPointerTable(rom, definition: definition)
        case "ROMProperty":         return ROMProperty(rom, definition: definition)
        case "ROMReference":        return ROMReference(rom, definition: definition)
        case "ROMScript":           return ROMScript(rom, definition: definition)
        case "ROMStringTable":      return ROMStringTable(rom, definition: definition)
        case "ROMText":             return ROMText(rom, definition: definition)
        case "ROMTextEncoding":     return ROMTextEncoding(rom, definition: definition)
        case "ROMTileLayout":       return ROMTileLayout(rom, definition: definition)
        default:                    return ROMObject(rom, definition: definition)
        }
    }
}
