//
//  ROMProperty.swift
//  ROMTools
//
//  Created by Evan Owen on 8/21/16.
//  Copyright © 2016 Evan Owen. All rights reserved.
//

import Foundation

class ROMProperty: ROMAssembly {
    
    // MARK: instance properties
    
    var property = 0
    let byteIndex: Int
    let bitMask: Int
    var bitIndex: Int
    override var description: String { return String(property) }
    
    // MARK: ROMAssembly methods
    
    override init(_ rom: ROM, definition: NSDictionary? = nil) {
        
        // get the byte index and bit mask
        byteIndex = Int(object: definition, keyPath: "byteIndex") ?? 0
        bitMask = Int(object: definition, keyPath: "bitMask") ?? 0xFF
        
        // get the bit index
        bitIndex = 0
        for bit in 0..<64 {
            if (((1 << bit) & bitMask) != 0) {
                bitIndex = bit
                break
            }
        }
//        bitIndex = 0
//        while (((1 << bitIndex) & bitMask) == 0) {
//            bitIndex += 1
//        }
        
        super.init(rom, definition: definition)
    }
    
    override func createDefinition() -> NSMutableDictionary {
        let definition = super.createDefinition()
        
        // type
        definition["type"] = "ROMProperty"
        
        // byte index
        if byteIndex != 0 {
            definition["byteIndex"] = byteIndex
        }
        
        // bit mask
        if bitMask != 0xFF {
            definition["bitMask"] = String(format: "0x%02X", bitMask)
        }
        
        return definition
    }
    
    // MARK: NSKeyValueCoding methods
    
    override func setNilValueForKey(_ key: String) {
        if key == "property" {
            property = 0
            return
        }
        super.setNilValueForKey(key)
    }
    
    // MARK: ROMAssembly methods
    
    override func assembleTo(_ data: inout Data) {
        var valueToInsert = 0
        var valueRange = NSMakeRange(byteIndex, 8)
        if data.count - byteIndex < 8 {
            valueRange.length = data.count - byteIndex
            if valueRange.length < 0 {
                rom.log("Unable to write value. Byte index larger than data size")
                return
            }
        }

        let oldData = NSMutableData(data: data)
        
        oldData.getBytes(&valueToInsert, range: valueRange)
        valueToInsert &= ~bitMask
        valueToInsert |= (property << bitIndex) & bitMask
        oldData.replaceBytes(in: valueRange, withBytes: &valueToInsert)
        data = oldData as Data
    }
    
    override func disassembleFrom(_ data: Data) {
        property = 0
        var valueRange = NSMakeRange(byteIndex, 8)
        if data.count - byteIndex < 8 {
            valueRange.length = data.count - byteIndex
            if valueRange.length < 0 {
                rom.log("Unable to read value. Byte index larger than data size")
                return
            }
        }
        (data as NSData).getBytes(&property, range: valueRange)
        property &= bitMask
        property >>= bitIndex
    }
}
